@@ -21,6 +21,7 @@ import serial
 from serial.tools import list_ports
 # noinspection PyPackageRequirementscd
 from serial.serialutil import SerialException
+import socket
 import sys
 import threading
 import time
@@ -43,14 +44,17 @@ class Pymata4(threading.Thread):
     def __init__(self, com_port=None, baud_rate=115200,
                  arduino_instance_id=1, arduino_wait=4,
                  sleep_tune=0.000001,
-                 shutdown_on_exception=True):
+                 shutdown_on_exception=True, ip_address=None,
+                 ip_port=None):
         """
         If you are using the Firmata Express Arduino sketch,
         and have a single Arduino connected to your computer,
         then you may accept all the default values.
 
         If you are using some other Firmata sketch, then
-        you must specify both the com_port and baudrate.
+        you must specify both the com_port and baudrate for
+        as serial connection, or ip_address and ip_port if
+        using StandardFirmataWifi.
 
         :param com_port: e.g. COM3 or /dev/ttyACM0.
 
@@ -69,7 +73,14 @@ class Pymata4(threading.Thread):
                                       a RunTimeError exception, or
                                       receiving a KeyboardInterrupt exception
 
+        :param ip_address: Used with StandardFirmataWifi to specify IP address of
+                           the WiFi device
+
+        :param ip_port: Used with StandardFirmataWifi to specify IP port of
+                           the WiFi device. Typically this is 3030
+
         """
+        self.start_time = time.time()
         # initialize threading parent
         threading.Thread.__init__(self)
 
@@ -80,9 +91,18 @@ class Pymata4(threading.Thread):
         self.the_reporter_thread = threading.Thread(target=self._reporter)
         self.the_reporter_thread.daemon = True
 
-        # create a thread to continuously receive serial data
-        self.the_serial_receive_thread = threading.Thread(target=self._serial_receiver)
-        self.the_serial_receive_thread.daemon = True
+        self.ip_address = ip_address
+        self.ip_port = ip_port
+
+        # if an ip address was specified, tcp/ip will be used instead of serial
+        # transfer.
+        # create a thread to continuously receive data
+        if self.ip_address:
+            self.the_data_receive_thread = threading.Thread(target=self._tcp_receiver)
+        else:
+            self.the_data_receive_thread = threading.Thread(target=self._serial_receiver)
+
+        self.the_data_receive_thread.daemon = True
 
         # keep alive variables
         self.keep_alive_interval = []
@@ -133,7 +153,8 @@ class Pymata4(threading.Thread):
         self.report_dispatch.update({PrivateConstants.I2C_REPLY: [self._i2c_reply, 2]})
         self.report_dispatch.update({PrivateConstants.CAPABILITY_RESPONSE: [self._capability_response, 2]})
         self.report_dispatch.update({PrivateConstants.PIN_STATE_RESPONSE: [self._pin_state_response, 2]})
-        self.report_dispatch.update({PrivateConstants.ANALOG_MAPPING_RESPONSE: [self._analog_mapping_response, 2]})
+        self.report_dispatch.update({PrivateConstants.ANALOG_MAPPING_RESPONSE: [self._analog_mapping_response, 4]})
+        self.report_dispatch.update({PrivateConstants.DHT_DATA: [self._dht_read_response, 12]})
 
         # report query results are stored in this dictionary
         self.query_reply_data = {PrivateConstants.REPORT_VERSION: '',
@@ -146,11 +167,17 @@ class Pymata4(threading.Thread):
 
         self.firmata_firmware = []
 
+        # dht error flag
+        self.dht_sensor_error = False
+
         # a list of PinData objects - one for each pin segregated by pin type
         # see pin_data.py
         self.analog_pins = []
         self.digital_pins = []
 
+        # a list of pins assigned to DHT devices
+        self.dht_list = []
+
         # This lock is used when the PinData object is update or contents
         # are retrieved
         self.the_pin_data_lock = threading.Lock()
@@ -167,6 +194,9 @@ class Pymata4(threading.Thread):
         # serial port in use
         self.serial_port = None
 
+        # handle to tcp/ip socket
+        self.sock = None
+
         # An i2c_map entry consists of a device i2c address as the key, and
         #  the value of the key consists of a dictionary containing 2 entries.
         #  The first entry. 'value' contains the last value reported, and
@@ -191,32 +221,42 @@ class Pymata4(threading.Thread):
 
         print(f"pymata4:  Version {PrivateConstants.PYMATA_EXPRESS_THREADED_VERSION}\n\n"
               f"Copyright (c) 2020 Alan Yorinks All Rights Reserved.\n")
-        if not self.com_port:
-            # user did not specify a com_port
-            try:
-                self._find_arduino()
-            except KeyboardInterrupt:
+        # if this is not a tcp interface, find the serial port
+        if not self.ip_address:
+            if not self.com_port:
+                # user did not specify a com_port
+                try:
+                    self._find_arduino()
+                except KeyboardInterrupt:
+                    if self.shutdown_on_exception:
+                        self.shutdown()
+            else:
+                # com_port specified - set com_port and baud rate
+                try:
+                    self._manual_open()
+                except KeyboardInterrupt:
+                    if self.shutdown_on_exception:
+                        self.shutdown()
+
+            if self.serial_port:
+                print(f"Arduino compatible device found and connected to {self.serial_port.port}")
+
+            # no com_port found - raise a runtime exception
+            else:
                 if self.shutdown_on_exception:
                     self.shutdown()
+                raise RuntimeError('No Arduino Found or User Aborted Program')
+        # this is tcp/ip interface
         else:
-            # com_port specified - set com_port and baud rate
-            try:
-                self._manual_open()
-            except KeyboardInterrupt:
-                if self.shutdown_on_exception:
-                    self.shutdown()
-
-        if self.serial_port:
-            print(f"Arduino compatible device found and connected to {self.serial_port.port}")
-
-        # no com_port found - raise a runtime exception
-        else:
-            if self.shutdown_on_exception:
-                self.shutdown()
-            raise RuntimeError('No Arduino Found or User Aborted Program')
+            # with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as self.sock:
+            #     s = self.sock.create_connection((self.ip_address, self.ip_port))
+            #     print(s)
+            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+            self.sock.connect((self.ip_address, self.ip_port))
+            print(f'Successfully connected to: {self.ip_address}:{self.ip_port}')
 
         self.the_reporter_thread.start()
-        self.the_serial_receive_thread.start()
+        self.the_data_receive_thread.start()
 
         # allow the threads to run
         self._run_threads()
@@ -273,6 +313,11 @@ class Pymata4(threading.Thread):
                 self.shutdown()
             raise RuntimeError('User Hit Control-C')
 
+        # Set the sampling interval to the standard value
+        # so the the DHT and HC-SRO4 device report at the right
+        # time frame.
+        self.set_sampling_interval(19)
+
     def _find_arduino(self):
         """
         This method will search all potential serial ports for an Arduino
@@ -389,6 +434,24 @@ class Pymata4(threading.Thread):
         """
         return self.analog_pins[pin].current_value, self.analog_pins[pin].event_time
 
+    def dht_read(self, pin):
+        """
+        Retrieve the last data update for the specified dht pin.
+
+        :param pin: digital pin number
+
+        :return: A list = [humidity, temperature  time_stamp]
+
+                 ERROR CODES: If either humidity or temperature value:
+                              == -1 Configuration Error
+                              == -2 Checksum Error
+                              == -3 Timeout Error
+
+        """
+        return self.digital_pins[pin].current_value[0], \
+            self.digital_pins[pin].current_value[1], \
+            self.digital_pins[pin].event_time
+
     def digital_read(self, pin):
         """
         Retrieve the last data update for the specified digital pin.
@@ -960,6 +1023,46 @@ class Pymata4(threading.Thread):
                            callback=callback,
                            differential=differential)
 
+    def set_pin_mode_dht(self, pin_number, sensor_type=22, differential=.1, callback=None):
+        """
+        Configure a DHT sensor prior to operation.
+        Up to 6 DHT sensors are supported
+
+        :param pin_number: digital pin number on arduino.
+
+        :param sensor_type: type of dht sensor
+                            Valid values = DHT11, DHT12, DHT22, DHT21, AM2301
+
+        :param differential: This value needs to be met for a callback
+                             to be invoked.
+
+        :param callback: callback function
+
+        callback: returns a data list:
+
+        [pin_type, pin_number, DHT type, humidity value, temperature raw_time_stamp]
+
+        The pin_type for DHT input pins = 15
+
+                ERROR CODES: If either humidity or temperature value:
+                              == -1 Configuration Error
+                              == -2 Checksum Error
+                              == -3 Timeout Error
+        """
+
+        # if the pin is not currently associated with a DHT device
+        # initialize it.
+        if pin_number not in self.dht_list:
+            self.dht_list.append(pin_number)
+            self.digital_pins[pin_number].cb = callback
+            self.digital_pins[pin_number].current_value = [0, 0]
+            self.digital_pins[pin_number].differential = differential
+            data = [pin_number, sensor_type]
+            self._send_sysex(PrivateConstants.DHT_CONFIG, data)
+        else:
+            # allow user to change the differential value
+            self.digital_pins[pin_number].differential = differential
+
     def set_pin_mode_digital_input(self, pin_number, callback=None):
         """
         Set a pin as a digital input.
@@ -1146,8 +1249,9 @@ class Pymata4(threading.Thread):
 
         :param pin_number: arduino pin number
 
-        :param pin_state: INPUT/OUTPUT/ANALOG/PWM/PULLUP - for SERVO use
-                          servo_config()
+        :param pin_state: INPUT/OUTPUT/ANALOG/PWM/PULLUP
+                         For SERVO use: set_pin_mode_servo
+                         For DHT   use: set_pin_mode_dht
 
         :param callback: A reference to a call back function to be
                          called when pin data value changes
@@ -1229,8 +1333,15 @@ class Pymata4(threading.Thread):
             for pin in range(len(self.digital_pins)):
                 self.disable_digital_reporting(pin)
             self.send_reset()
-            self.serial_port.reset_input_buffer()
-            self.serial_port.close()
+            if self.ip_address:
+                try:
+                    self.sock.shutdown(socket.SHUT_RDWR)
+                    self.sock.close()
+                except Exception:
+                    pass
+            else:
+                self.serial_port.reset_input_buffer()
+                self.serial_port.close()
 
         except (RuntimeError, SerialException, OSError):
             # ignore error on shutdown
@@ -1257,7 +1368,7 @@ class Pymata4(threading.Thread):
         if sonar_pin_entry:
             return [sonar_pin_entry[1], sonar_pin_entry[2]]
         else:
-            return[0, 0]
+            return [0, 0]
 
     def stepper_write(self, motor_speed, number_of_steps):
         """
@@ -1330,6 +1441,102 @@ class Pymata4(threading.Thread):
         """
         self.query_reply_data[PrivateConstants.CAPABILITY_RESPONSE] = data
 
+    def _dht_read_response(self, data):
+        """
+        Process the dht response message.
+
+        Values are calculated using:
+                humidity = (_bits[0] * 256 + _bits[1]) * 0.1
+
+                temperature = ((_bits[2] & 0x7F) * 256 + _bits[3]) * 0.1
+
+        error codes:
+        0 - OK
+        1 - DHTLIB_ERROR_TIMEOUT
+        2 - Checksum error
+
+        :param: data - array of 9 7bit bytes ending with the error status
+        """
+        # get the time of the report
+        time_stamp = time.time()
+        # initiate a list for a potential call back
+        reply_data = [PrivateConstants.DHT]
+
+        # get the pin and type of the dht
+        pin = data[0]
+        reply_data.append(pin)
+        dht_type = data[1]
+        reply_data.append(dht_type)
+        humidity = None
+        temperature = None
+
+        self.digital_pins[pin].event_time = time_stamp
+
+        if data[11] == 1:  # data[9] is config flag
+            if data[10] != 0:
+                self.dht_sensor_error = True
+                humidity = temperature = -1
+                # return
+        else:
+            # if data read correctly process and return
+
+            if data[10] == 0:
+                # dht 22
+                if data[1] == 22:
+                    humidity = (((data[2] & 0x7f) + (data[3] << 7)) * 256 +
+                                ((data[4] & 0x7f) + (data[5] << 7))) * 0.1
+                    temperature = (((data[6] & 0x7f) + (data[7] << 7) & 0x7F) * 256 +
+                                   ((data[8] & 0x7f) + (data[9] << 7))) * 0.1
+                # dht 11
+                elif data[2] == 11:
+                    humidity = (((data[2] & 0x7f) + (data[3] << 7)) +
+                                ((data[4] & 0x7f) + (data[5] << 7))) * 0.1
+                    temperature = (((data[6] & 0x7f) + (data[7] << 7) & 0x7F) +
+                                   ((data[8] & 0x7f) + (data[9] << 7))) * 0.1
+                else:
+                    raise RuntimeError(f'Unknown DHT Sensor type reported: {data[2]}')
+
+                humidity = round(humidity, 2)
+                temperature = round(temperature, 2)
+
+                # check for negative temperature
+                if data[6] & 0x80:
+                    temperature = -temperature
+
+            elif data[8] == 1:
+                # Checksum Error
+                humidity = temperature = -2
+                self.dht_sensor_error = True
+            elif data[8] == 2:
+                # Timeout Error
+                humidity = temperature = -3
+                self.dht_sensor_error = True
+        # since we initialize
+        if humidity is None:
+            return
+        reply_data.append(humidity)
+        reply_data.append(temperature)
+        reply_data.append(time_stamp)
+
+        # retrieve the last reported values
+        last_value = self.digital_pins[pin].current_value
+
+        self.digital_pins[pin].current_value = [humidity, temperature]
+        if self.digital_pins[pin].cb:
+            # only report changes
+            # has the humidity changed?
+            if last_value[0] != humidity:
+
+                differential = abs(humidity - last_value[0])
+                if differential >= self.digital_pins[pin].differential:
+                    self.digital_pins[pin].cb(reply_data)
+                return
+            if last_value[1] != temperature:
+                differential = abs(temperature - last_value[1])
+                if differential >= self.digital_pins[pin].differential:
+                    self.digital_pins[pin].cb(reply_data)
+                return
+
     def _digital_message(self, data):
         """
         This is a private message handler method.
@@ -1494,13 +1701,16 @@ class Pymata4(threading.Thread):
         """
         # send_message = ""
         send_message = bytes(command)
-        try:
-            result = self.serial_port.write(send_message)
-        except SerialException:
-            if self.shutdown_on_exception:
-                self.shutdown()
-            raise RuntimeError('write fail in _send_command')
-        return result
+        if not self.ip_address:
+            try:
+                result = self.serial_port.write(send_message)
+            except SerialException:
+                if self.shutdown_on_exception:
+                    self.shutdown()
+                raise RuntimeError('write fail in _send_command')
+            return result
+        else:
+            self.sock.sendall(send_message)
 
     def _send_keep_alive(self):
         """
@@ -1709,3 +1919,16 @@ class Pymata4(threading.Thread):
                     # continue
             except OSError:
                 pass
+
+    def _tcp_receiver(self):
+        """
+        Thread to continuously check for incoming data.
+        When a byte comes in, place it onto the deque.
+        """
+        self.run_event.wait()
+        while self._is_running() and not self.shutdown_flag:
+            try:
+                payload = self.sock.recv(1)
+                self.the_deque.append(ord(payload))
+            except Exception:
+                pass
